// include local
#include "TreeWriter.hpp"


//ClassImp(TreeWriter)


//**************************************************************************************************************************
TreeWriter::TreeWriter()
  : m_isFirst(true), m_counter(0), m_file( new TFile( "output.root", "RECREATE" ) ), m_tree( new TTree( "data", "data" ) ), m_temp(0) {
}


//**************************************************************************************************************************
TreeWriter::TreeWriter( const std::string& filename, const std::string& treename )
  : m_isFirst(true), m_counter(0), m_file( new TFile( filename.c_str(), "RECREATE" ) ), m_tree( new TTree( treename.c_str(), treename.c_str() ) ), m_temp(0) {

}


//**************************************************************************************************************************
TreeWriter::TreeWriter( const std::string& filename, TTree* tree ) 
  : m_isFirst(true), m_counter(0), m_file( new TFile( filename.c_str(), "RECREATE" ) ) {

  m_tree =  dynamic_cast< TTree* >( tree->CloneTree(0) );
  m_tree->SetName( tree->GetName() );
  //m_tree->SetBranchStatus("*",1); // select all branches
  m_temp = tree;
}


//**************************************************************************************************************************
TreeWriter::~TreeWriter() {
 
  m_file->cd();
  m_file->ls();
  m_tree->ResetBranchAddresses();
  //m_tree->Write(); 
  m_file->Write();
  m_file->Close();
  delete m_file; m_file = 0;
}


//**************************************************************************************************************************
void TreeWriter::write( ) throw( TreeWriterException ) {
    
  // Check that the number of attempted variable writes is the same as the last
  if( m_variables.size() != m_counter ) {
    throw TreeWriterException( "Not enough variables have been added compared with previous write" );
  }
  
  // Fill the tree with data
  m_tree->Fill();
  if( m_isFirst ) {
     m_isFirst = false;
  }

  m_counter = 0;

}


//**************************************************************************************************************************
void TreeWriter::write( const unsigned int& entry ) throw( TreeWriterException ) {

  // Check that the number of attempted variable writes is the same as the last
  if( m_variables.size() != m_counter ) {
    throw TreeWriterException( "Not enough variables have been added compared with previous write" );
  }

  // Fill the tree with data  
  if ( m_temp != 0 ) {
    m_temp->GetEntry( entry );
  }

  m_tree->Fill();

  if( m_isFirst ) {
     m_isFirst = false;
  }
  m_counter = 0;
}


//**************************************************************************************************************************
void TreeWriter::column( const std::string& variable, const boost::any& value ) throw( TreeWriterException ) {

  if( m_isFirst ) {

    m_variables.push_back( variable );
    check_variable( variable, value );
  } else {

    // check that the variable is actually in the map
    if( std::find( m_variables.begin(), m_variables.end(), variable ) != m_variables.end() ) {
      
      check_variable( variable, value );

    } else {

      TreeWriterException( "WARNING - Variable " + variable + " has not been added to tuple." );

    }
  }
}


//**************************************************************************************************************************
void TreeWriter::check_variable( const std::string& variable, const boost::any& value ) {

  try {

    if( is_empty( value ) ){

      throw TreeWriterException( "Value is empty" );

    } else if( is_int( value ) ) {

      if( m_int_variables.find( variable ) == m_int_variables.end() ) {

        m_int_variables.insert( std::make_pair( variable, 0 ) );
        m_tree->Branch( variable.c_str(), &m_int_variables[ variable ] );
      
      } 

      m_int_variables[variable] = boost::any_cast<Int_t>( value );
      m_counter++;

    } else if( is_float( value ) ) {

      if( m_float_variables.find( variable ) == m_float_variables.end() ) {

        m_float_variables.insert( std::make_pair( variable, 0 ) );
        m_tree->Branch( variable.c_str(), &m_float_variables[ variable ] );
      
      } 

      m_float_variables[variable] = boost::any_cast<Float_t>( value );
      m_counter++;

    } else if( is_double( value ) ) {

      if( m_double_variables.find( variable ) == m_double_variables.end() ) {

        m_double_variables.insert( std::make_pair( variable, 0 ) );
        m_tree->Branch( variable.c_str(), &m_double_variables[ variable ] );
      
      } 
      
      m_double_variables[variable] = boost::any_cast<Double_t>( value );
      m_counter++;

    } else if( is_bool( value ) ) {

      if( m_bool_variables.find( variable ) == m_bool_variables.end() ) {

        m_bool_variables.insert( std::make_pair( variable, 0 ) );
        m_tree->Branch( variable.c_str(), &m_bool_variables[ variable ] );
      
      } 

      m_bool_variables[variable] = boost::any_cast<Bool_t>( value );
      m_counter++;

    }  else {
  
  	std::cerr << "WARNING: Variable " << variable << " has not been added to tuple." << std::endl;
      std::cerr << "VALUE NOT RECOGNISED" << std::endl;
    }

  } catch( const boost::bad_any_cast& e ) {
  //} catch( const std::exception& e ) {

    throw TreeWriterException( e.what() );
  }

}


//**************************************************************************************************************************
void TreeWriter::setTreeName( const std::string& entry ) {

  setTreeName( entry.c_str() );
}


//**************************************************************************************************************************
void TreeWriter::setTreeName( const char* entry ) {

  m_tree->SetName( entry );
}


//**************************************************************************************************************************
bool TreeWriter::is_empty( const boost::any& operand ) {
    return operand.empty();
}


//**************************************************************************************************************************
bool TreeWriter::is_int( const boost::any& operand ) {
    return operand.type() == typeid(Int_t);
}


//**************************************************************************************************************************
bool TreeWriter::is_float( const boost::any& operand ) {
    return operand.type() == typeid(Float_t);
}


//**************************************************************************************************************************
bool TreeWriter::is_double( const boost::any& operand ) {
    return operand.type() == typeid(Double_t);
}


//**************************************************************************************************************************
bool TreeWriter::is_bool( const boost::any& operand ) {
    return operand.type() == typeid(Bool_t);
}


